#include <stdio.h>
#include <string.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>
#include <gmpc/plugin.h>
#include <gmpc/gmpc_easy_download.h>
#include <gmpc/metadata.h>
#include <gmpc/gmpc-mpddata-model.h>
#include <gmpc/gmpc-mpddata-treeview.h>
#include <gmpc/TreeSearchWidget.h>
#include <config.h>
#include "magnatune.h"

extern GladeXML *pl3_xml;

gmpcPlugin plugin;
/**
 * Function decleration
 */
static void magnatune_selected(GtkWidget *container);
static void magnatune_unselected(GtkWidget *container);
static void magnatune_add(GtkWidget *cat_tree);
static void magnatune_download_xml_callback(int download, int total,gpointer data);
static void magnatune_download();
static void   magnatune_mpd_status_changed(MpdObj *mi, ChangedStatusType what, void *data);
static int magnatune_fetch_cover_priority(void);
static int magnatune_fetch_get_image(mpd_Song *song,MetaDataType type, char **path);
static int magnatune_cat_menu_popup(GtkWidget *menu, int type, GtkWidget *tree, GdkEventButton *event);

static int magnatune_button_release_event(GtkWidget *tree, GdkEventButton *event,gpointer data);
static int magnatune_button_handle_release_event_tag(GtkWidget *tree, GdkEventButton *event,gpointer data);

static void magnatune_cat_key_press(GtkWidget *tree, GdkEventKey *event, int selected_type);
static int magnatune_key_press(GtkWidget *tree, GdkEventKey *event);
static void magnatune_show_album_list(GtkTreeSelection *selection, gpointer user_data);
static void magnatune_show_song_list(GtkTreeSelection *selection, gpointer user_data);
static int magnatune_button_handle_release_event_tag_add(GtkWidget *button, gpointer user_data);
static void magnatune_save_myself(void);
static GtkTreeRowReference *magnatune_ref = NULL;
/**
 * Get set enabled 
 */
static int magnatune_get_enabled()
{
	return cfg_get_single_value_as_int_with_default(config, "magnatune", "enable", TRUE);
}
static void magnatune_set_enabled(int enabled)
{
	cfg_set_single_value_as_int(config, "magnatune", "enable", enabled);
	if (enabled)
	{
		if(magnatune_ref == NULL)
		{
			magnatune_add(GTK_WIDGET(playlist3_get_category_tree_view()));
		}
	}
	else if (magnatune_ref)
	{
		GtkTreePath *path = gtk_tree_row_reference_get_path(magnatune_ref);
		if (path){
			GtkTreeIter iter;
            magnatune_save_myself();
			if (gtk_tree_model_get_iter(GTK_TREE_MODEL(playlist3_get_category_tree_store()), &iter, path)){
				gtk_tree_store_remove(playlist3_get_category_tree_store(), &iter);
			}
			gtk_tree_path_free(path);
			gtk_tree_row_reference_free(magnatune_ref);
			magnatune_ref = NULL;
		}                                                                                                  	
	}                                                                                                      	
	pl3_update_go_menu();
}





/* Playlist window row reference */

GtkWidget *magnatune_pb = NULL;
GtkWidget *magnatune_vbox = NULL;
GtkTreeModel *mt_store = NULL;
TreeSearch *mt_tree_search = NULL;

GtkWidget *treeviews[3] = {NULL, NULL,NULL};

static GtkWidget *magnatune_logo=NULL;
extern GladeXML *pl3_xml;

static void magnatune_buy_album()
{
	gchar *uri = "http://www.magnatune.com/";
	gchar *path = g_strdup_printf("%s%c%s%c%s '%s'",plugin.path,G_DIR_SEPARATOR, "magnatune",G_DIR_SEPARATOR,"xdg-open",uri);
	g_spawn_command_line_async(path, NULL);
	g_free(path);
}

static void magnatune_logo_add()
{
	mpd_Song *song = mpd_playlist_get_current_song(connection);
	GtkWidget *logo, *button;
	GtkWidget *ali;
	
	magnatune_logo = gtk_hbox_new(FALSE,6);
	
	button = gtk_button_new_with_label("Buy this album\nfrom magnatune");
	gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);

	ali = gtk_alignment_new(0,0.5,0,1);
	gtk_container_add(GTK_CONTAINER(ali), button);
	logo = gtk_image_new_from_icon_name("magnatune", GTK_ICON_SIZE_DND);
	gtk_button_set_image(GTK_BUTTON(button), logo);
	gtk_box_pack_start(GTK_BOX(magnatune_logo), ali, TRUE, TRUE,0);
	gtk_box_pack_end(GTK_BOX(glade_xml_get_widget(pl3_xml, "vbox5")), magnatune_logo, FALSE,FALSE,0);	
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(magnatune_buy_album), NULL);
	if(song) {
		if(strstr(song->file,"magnatune.com"))
		{
			gtk_widget_show_all(magnatune_logo);
			return;
		}
	}


}

static void magnatune_logo_init()
{
	gchar *path = NULL;

	path = g_strdup_printf("%s%c%s",plugin.path,G_DIR_SEPARATOR, "magnatune");
	gtk_icon_theme_append_search_path(gtk_icon_theme_get_default (),path);
	g_free(path);

	gtk_init_add((GtkFunction )magnatune_logo_add, NULL);
	/** 
	 * init the magnatune stuff
	 */
	magnatune_db_init();
	/**
	 * open the db 
	 */
	magnatune_db_open();
}


static void   magnatune_mpd_status_changed(MpdObj *mi, ChangedStatusType what, void *data)
{
	if(!magnatune_logo)
		return;
	if(what&(MPD_CST_SONGID|MPD_CST_STATE))
	{
		mpd_Song *song = mpd_playlist_get_current_song(mi);
		if(song) {
			if(strstr(song->file,"magnatune.com"))
			{
				gtk_widget_show_all(magnatune_logo);
				return;
			}
		}
		gtk_widget_hide(magnatune_logo);
	}
}

static void magnatune_get_genre_list()
{
    MpdData *data = NULL;
	data = magnatune_db_get_genre_list();
    gmpc_mpddata_model_set_mpd_data(GMPC_MPDDATA_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[0]))),data);
}

static void magnatune_show_album_list(GtkTreeSelection *selection, gpointer user_data)
{
    MpdData *data = NULL;
    GtkTreeIter piter;
    GtkTreeModel *model  = gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[0]));
    if(gtk_tree_selection_get_selected(selection, &model, &piter))
    {
        gchar *genre;
        gtk_tree_model_get(model, &piter, MPDDATA_MODEL_COL_SONG_TITLE,&genre,-1);
        data= magnatune_db_get_artist_list(genre);
        
        g_free(genre);
    }
    gmpc_mpddata_model_set_mpd_data(GMPC_MPDDATA_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[1]))),data);
}

static void magnatune_show_song_list(GtkTreeSelection *selection, gpointer user_data)
{
    MpdData *data = NULL;
    GtkTreeIter piter;
    GtkTreeSelection *select = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[0]));
    GtkTreeModel *model  = gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[0]));
    gchar *genre = NULL, *artist = NULL, *album = NULL;
    if(gtk_tree_selection_get_selected(select, &model, &piter))
    {

        select = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[1]));
        gtk_tree_model_get(model, &piter, MPDDATA_MODEL_COL_SONG_TITLE,&genre,-1);
        /* model */
        model  = gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[1]));
        if(gtk_tree_selection_get_selected(select, &model, &piter))
        {
            gtk_tree_model_get(model, &piter, MPDDATA_MODEL_COL_SONG_TITLE,&artist,-1);
            select = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[2]));
            model  = gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[2]));
            if(gtk_tree_selection_get_selected(select, &model, &piter))
            {
                gtk_tree_model_get(model, &piter, MPDDATA_MODEL_COL_SONG_TITLE,&album,-1);
            }
        }
    }
    data = magnatune_db_get_song_list(genre, artist, album);
    gmpc_mpddata_model_set_mpd_data(GMPC_MPDDATA_MODEL(mt_store), data);
}

static void magnatune_show_artist_list(GtkTreeSelection *selection, gpointer user_data)
{
    GtkTreeIter piter;
    MpdData *data = NULL;
    GtkTreeSelection *select = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[0]));
    GtkTreeModel *model  = gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[0]));

    if(gtk_tree_selection_get_selected(select, &model, &piter))
    {
        GList *iter,*list = NULL;
        gchar *genre;

        select = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[1]));
        gtk_tree_model_get(model, &piter, MPDDATA_MODEL_COL_SONG_TITLE,&genre,-1);
        /* model */
        model  = gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[1]));
        if(gtk_tree_selection_get_selected(select, &model, &piter))
        {
            GList *iter,*list = NULL;

            gchar *artist;
            gtk_tree_model_get(model, &piter, MPDDATA_MODEL_COL_SONG_TITLE,&artist,-1);
            data = magnatune_db_get_album_list(genre ,artist);
            gmpc_mpddata_model_set_request_artist(GMPC_MPDDATA_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[2]))),artist);
            g_free(artist);
        }
        g_free(genre);
    }
    gmpc_mpddata_model_set_mpd_data(GMPC_MPDDATA_MODEL(gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[2]))),data);
}

static void magnatune_add_album_row_activated(GtkTreeView *tree, GtkTreePath *path, GtkTreeViewColumn *col, gpointer data)
{
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GtkTreeIter iter_r;

    if(gtk_tree_model_get_iter(model, &iter_r, path))
    {
        gchar *path;
        gtk_tree_model_get(model, &iter_r, MPDDATA_MODEL_COL_PATH, &path, -1);
        play_path(path);
        g_free(path);
    }
}

static gboolean magnatune_search_equal_func(GtkTreeModel *model, int column, const gchar *key, GtkTreeIter *iter, gpointer data)
{
	gchar *compare = NULL;
	gtk_tree_model_get(model,iter, column, &compare, -1);
	if(compare){
		gchar *a = g_utf8_casefold(key, -1);
		gchar *b = g_utf8_casefold(compare, -1);
		if(strstr(b,a))
		{
			g_free(a);
			g_free(b);
			return FALSE;
		}
		g_free(a);
		g_free(b);
	}
	return TRUE;
}

static void magnatune_search_activate(TreeSearch *ts)
{
	GtkTreeModel *model = GTK_TREE_MODEL(mt_store);
	GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(ts->treeview));
	if (gtk_tree_selection_count_selected_rows (selection) == 1)            
	{
		GList *list = gtk_tree_selection_get_selected_rows (selection, &model);
		magnatune_add_album_row_activated(GTK_TREE_VIEW(ts->treeview),(GtkTreePath *)list->data, NULL,NULL);	
		/* free list */
		g_list_foreach (list, (GFunc) gtk_tree_path_free, NULL);                        	
		g_list_free (list);
	}
}

static void magnatune_init()
{
    int size;
	GtkWidget *tree             = NULL;
	GtkWidget *sw               = NULL;
	GtkCellRenderer *renderer   = NULL;
    GtkWidget *paned            = NULL;
    GtkWidget *hbox             = NULL;
    GtkWidget *vbox             = NULL;
    GtkTreeModel *model         = NULL;
    GtkTreeViewColumn *column   = NULL;
    int position  = cfg_get_single_value_as_int_with_default(config, "magnatune", "pane-pos", 150);
	magnatune_vbox = gtk_vpaned_new();
    gtk_paned_set_position(GTK_PANED(magnatune_vbox), position);
   

    vbox = gtk_vbox_new(FALSE,6);
	/**
	 * Create list store
	 * 1. pointer 
	 * 2. name
	 */
	mt_store =  (GtkTreeModel *)gmpc_mpddata_model_new(); 

    /* Paned window */
    paned = gtk_hbox_new(TRUE,6);
    model = (GtkTreeModel *) gmpc_mpddata_model_new();
    /* Genre list */
    sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    treeviews[0] = gtk_tree_view_new_with_model(model);
    gtk_tree_view_set_search_column(GTK_TREE_VIEW(treeviews[0]), MPDDATA_MODEL_COL_SONG_TITLE);
    g_signal_connect(G_OBJECT(treeviews[0]), "button-press-event", G_CALLBACK(magnatune_button_handle_release_event_tag), GINT_TO_POINTER(0));
    /* Add column */
    column = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title(column, "Genre");


	renderer = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
    gtk_tree_view_column_add_attribute(column, renderer, "icon-name", MPDDATA_MODEL_COL_ICON_ID);
    gtk_tree_view_column_set_sizing(column , GTK_TREE_VIEW_COLUMN_AUTOSIZE);
	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, TRUE);
	gtk_tree_view_column_add_attribute(column, renderer, "text", MPDDATA_MODEL_COL_SONG_TITLE);
	gtk_tree_view_insert_column(GTK_TREE_VIEW(treeviews[0]),column, -1);

    gtk_container_add(GTK_CONTAINER(sw), treeviews[0]);
    gtk_box_pack_start(GTK_BOX(paned), sw, TRUE, TRUE, 0); 

    g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[0]))), "changed", G_CALLBACK(magnatune_show_album_list), NULL);

    /* Artist list */
    model = (GtkTreeModel *) gmpc_mpddata_model_new();
    sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    treeviews[1] =  gtk_tree_view_new_with_model(model);

    gtk_tree_view_set_search_column(GTK_TREE_VIEW(treeviews[1]), MPDDATA_MODEL_COL_SONG_TITLE);
	g_signal_connect(G_OBJECT(treeviews[1]), "button-press-event", G_CALLBACK(magnatune_button_handle_release_event_tag), GINT_TO_POINTER(1));
    /* Add column */
    column = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title(column, "Artist");
    size = cfg_get_single_value_as_int_with_default(config, "gmpc-mpddata-model", "icon-size", 64);
    gtk_tree_view_column_set_sizing(column , GTK_TREE_VIEW_COLUMN_FIXED);
    gtk_tree_view_set_fixed_height_mode(GTK_TREE_VIEW(treeviews[1]), TRUE);


    renderer = gtk_cell_renderer_pixbuf_new();
    gtk_cell_renderer_set_fixed_size(renderer, size,size);
   
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
    gtk_tree_view_column_add_attribute(column, renderer, "pixbuf", MPDDATA_MODEL_META_DATA);
	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, TRUE);
	gtk_tree_view_column_add_attribute(column, renderer, "text", MPDDATA_MODEL_COL_SONG_TITLE);
	gtk_tree_view_insert_column(GTK_TREE_VIEW(treeviews[1]),column, -1);

    gtk_container_add(GTK_CONTAINER(sw), treeviews[1]);
    gtk_box_pack_start(GTK_BOX(paned), sw, TRUE, TRUE, 0); 

    g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[1]))), "changed", G_CALLBACK(magnatune_show_artist_list), NULL);

    /* Album list */
    model = (GtkTreeModel *) gmpc_mpddata_model_new();
    sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    treeviews[2] =  gtk_tree_view_new_with_model(model);
    gtk_tree_view_set_search_column(GTK_TREE_VIEW(treeviews[2]), MPDDATA_MODEL_COL_SONG_TITLE);

	g_signal_connect(G_OBJECT(treeviews[2]), "button-press-event", G_CALLBACK(magnatune_button_handle_release_event_tag), GINT_TO_POINTER(2));
    /* Add column */
    column = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title(column, "Album");
    size = cfg_get_single_value_as_int_with_default(config, "gmpc-mpddata-model", "icon-size", 64);
    gtk_tree_view_column_set_sizing(column , GTK_TREE_VIEW_COLUMN_FIXED);
    gtk_tree_view_set_fixed_height_mode(GTK_TREE_VIEW(treeviews[2]), TRUE);


	renderer = gtk_cell_renderer_pixbuf_new();
    gtk_cell_renderer_set_fixed_size(renderer, size,size);

	gtk_tree_view_column_pack_start(column, renderer, FALSE);
    gtk_tree_view_column_add_attribute(column, renderer, "pixbuf", MPDDATA_MODEL_META_DATA);
	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, TRUE);
	gtk_tree_view_column_add_attribute(column, renderer, "text", MPDDATA_MODEL_COL_SONG_TITLE);
	gtk_tree_view_insert_column(GTK_TREE_VIEW(treeviews[2]),column, -1);

    gtk_container_add(GTK_CONTAINER(sw), treeviews[2]);
    gtk_box_pack_start(GTK_BOX(paned), sw, TRUE, TRUE, 0); 


    g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[0]))), "changed", G_CALLBACK(magnatune_show_song_list), NULL);
    g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[1]))), "changed", G_CALLBACK(magnatune_show_song_list), NULL);
    g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[2]))), "changed", G_CALLBACK(magnatune_show_song_list), NULL);

    gtk_paned_add1(GTK_PANED(magnatune_vbox), paned);
    gtk_widget_show_all(paned);
    /**
	 * tree
	 */
    sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
 

	tree = (GtkWidget *)gmpc_mpddata_treeview_new("magnatune",TRUE,GTK_TREE_MODEL(mt_store));
	g_signal_connect(G_OBJECT(tree), "row-activated", G_CALLBACK(magnatune_add_album_row_activated), NULL);

	g_signal_connect(G_OBJECT(tree), "button-press-event", G_CALLBACK(magnatune_button_release_event), tree);
	g_signal_connect(G_OBJECT(tree), "key-press-event", G_CALLBACK(magnatune_key_press), NULL);

    gtk_container_add(GTK_CONTAINER(sw), tree);
    gtk_box_pack_start(GTK_BOX(vbox),sw, TRUE, TRUE,0);	

	mt_tree_search = (TreeSearch *)treesearch_new(GTK_TREE_VIEW(tree), MPDDATA_MODEL_COL_SONG_TITLE);
	gtk_box_pack_end(GTK_BOX(vbox), GTK_WIDGET(mt_tree_search), FALSE, TRUE, 0);	
	g_signal_connect(G_OBJECT(mt_tree_search), "result-activate", G_CALLBACK(magnatune_search_activate),NULL);


	gtk_widget_show_all(sw);
    gtk_widget_show(vbox);
	/**
	 * Progress Bar for the bottom 
	 */
	magnatune_pb = gtk_progress_bar_new();
	gtk_box_pack_end(GTK_BOX(vbox), magnatune_pb, FALSE, TRUE, 0);
    gtk_paned_add2(GTK_PANED(magnatune_vbox), vbox);


	g_object_ref(magnatune_vbox);


}
static void magnatune_selected(GtkWidget *container)
{
	if(magnatune_vbox == NULL)
	{
		magnatune_init();
		gtk_container_add(GTK_CONTAINER(container), magnatune_vbox);
		gtk_widget_show(magnatune_vbox);
		if(!magnatune_db_has_data())
		{
			magnatune_download();
		}
		magnatune_get_genre_list();
	} else {
		gtk_container_add(GTK_CONTAINER(container), magnatune_vbox);
		gtk_widget_show(magnatune_vbox);
	}
}
static void magnatune_unselected(GtkWidget *container)
{
	gtk_container_remove(GTK_CONTAINER(container), magnatune_vbox);
}

static void magnatune_download_xml_callback(int download, int total,gpointer data)
{
	GtkWidget *pb = data;
	gchar *label = NULL;
	if(total > 0)
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pb), download/(float)total);
	else
		gtk_progress_bar_pulse(GTK_PROGRESS_BAR(pb));
	if(download >> 20)
	{
		label = g_strdup_printf("%.2f mb",(download>>10)/1024.0);
	}
	else if (download>>10)
	{
		label = g_strdup_printf("%i kb",download>>10);
	}
	else
	{
		label = g_strdup_printf("%i b",download);
	}
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(pb), label);
	g_free(label);
	while(gtk_events_pending())
		gtk_main_iteration();
}

static void magnatune_download()
{
	gtk_widget_set_sensitive(glade_xml_get_widget(pl3_xml, "pl3_win"), FALSE);
	gtk_widget_show(magnatune_pb);
	magnatune_db_download_xml(magnatune_download_xml_callback, magnatune_pb);
	gtk_widget_hide(magnatune_pb);
	gtk_widget_set_sensitive(glade_xml_get_widget(pl3_xml, "pl3_win"), TRUE);
}

static void magnatune_add(GtkWidget *cat_tree)
{
	GtkTreePath *path = NULL;
	GtkTreeIter iter,child;
	GtkTreeStore *pl3_tree = (GtkTreeStore *)gtk_tree_view_get_model(GTK_TREE_VIEW(cat_tree));	
	gint pos = cfg_get_single_value_as_int_with_default(config, "magnatune","position",20);

	if(!cfg_get_single_value_as_int_with_default(config, "magnatune", "enable", TRUE)) return;

    debug_printf(DEBUG_INFO,"Adding at position: %i", pos);
	playlist3_insert_browser(&iter, pos);
	/*gtk_tree_store_append(pl3_tree, &iter, NULL);*/
	gtk_tree_store_set(pl3_tree, &iter, 
			PL3_CAT_TYPE, plugin.id,
			PL3_CAT_TITLE, "Magnatune Browser",
			PL3_CAT_INT_ID, "",
			PL3_CAT_ICON_ID, "magnatune",
			PL3_CAT_PROC, TRUE,
			PL3_CAT_ICON_SIZE,GTK_ICON_SIZE_DND,-1);
	/**
	 * Clean up old row reference if it exists
	 */
	if (magnatune_ref)
	{
		gtk_tree_row_reference_free(magnatune_ref);
		magnatune_ref = NULL;
	}
	/**
	 * create row reference
	 */
	path = gtk_tree_model_get_path(GTK_TREE_MODEL(playlist3_get_category_tree_store()), &iter);
	if (path)
	{
		magnatune_ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(playlist3_get_category_tree_store()), path);
		gtk_tree_path_free(path);
	}

}
static char * __magnatune_process_string(char *name)
{
	int i = 0;
	int j = 0;
	int depth = 0;
	int spaces = 0;

	/* only gonna be smaller */
	char *result = g_malloc0((strlen(name)+1+spaces*2)*sizeof(char));
	for(i=0; i < strlen(name);i++)
	{
		if(name[i] == '(' || name[i] == '[') depth++;
		else if (name[i] == ')' || name[i] == ']') depth--;
		else if(!depth)
		{
			result[j] = name[i];
			j++;
		}
	}
	for(i=j-1;i>0 && result[i] == ' ';i--)
	{
		result[i] = '\0';
	}
	return result;
}

/*** COVER STUFF */
static int magnatune_fetch_cover_priority(void){
	return cfg_get_single_value_as_int_with_default(config, "magnatune", "priority", 80);
}
static int magnatune_fetch_get_image(mpd_Song *song,MetaDataType type, char **path)
{
	int result = 0;
	if(magnatune_db_has_data() == FALSE)
		return META_DATA_UNAVAILABLE;
	if(type == META_ARTIST_ART && song->artist)
	{
		char *artist = __magnatune_process_string(song->artist);
		char *value = magnatune_db_get_value(artist,NULL,META_ARTIST_ART);
		if(value)
		{
			gchar *filename = NULL;
			gmpc_easy_download_struct dld = {NULL, 0, -1, NULL,NULL};
			filename = g_strdup_printf("%s.jpg", artist);
			*path = gmpc_get_covers_path(filename);
			g_free(filename); 
			if(gmpc_easy_download(value, &dld))
			{
				g_file_set_contents(*path, dld.data, dld.size, NULL);
				return META_DATA_AVAILABLE;	
			}
                        gmpc_easy_download_clean(&dld);
			g_free(*path);
			*path = NULL;
		}
		g_free(artist);
	}
	else if(type == META_ALBUM_ART && song->artist && song->album)
	{
		char *artist = __magnatune_process_string(song->artist);
		char *album= __magnatune_process_string(song->album);
		char *value = magnatune_db_get_value(artist,album,META_ALBUM_ART);
		if(value)
		{
			gchar *filename;
			gmpc_easy_download_struct dld = {NULL, 0, -1, NULL,NULL};
			filename = g_strdup_printf("%s-%s.jpg", artist,album);
			*path = gmpc_get_covers_path(filename);
			g_free(filename); 
			if(gmpc_easy_download(value, &dld))
			{
				g_file_set_contents(*path, dld.data, dld.size, NULL);
				return META_DATA_AVAILABLE;	
			}
                        gmpc_easy_download_clean(&dld);
			g_free(*path);
			*path = NULL;
		}
		g_free(artist);
	}
	return META_DATA_UNAVAILABLE; 
}	
static void magnatune_redownload_reload_db()
{
	GtkTreeIter iter;
	GtkTreeModel *model = gtk_tree_row_reference_get_model(magnatune_ref);
	GtkTreePath *path = gtk_tree_row_reference_get_path(magnatune_ref);
	if(path && gtk_tree_model_get_iter(model, &iter, path))
	{
		GtkTreeIter citer;
		while(gtk_tree_model_iter_children(model, &citer,&iter))
		{
			gtk_tree_store_remove(GTK_TREE_STORE(model), &citer);
		}
		magnatune_download();
		magnatune_get_genre_list();
	}
	if(path)
		gtk_tree_path_free(path);
}

static void magnatune_cat_menu_add()
{
	GtkTreeIter citer;
	GtkTreeSelection *sel = gtk_tree_view_get_selection(playlist3_get_category_tree_view());
	GtkTreeModel *model = (GtkTreeModel *)playlist3_get_category_tree_store();
	GtkTreePath *tpath;
	int depth = 0;
	if(!gtk_tree_selection_get_selected(sel,&model,&citer))
	{
		return ;
	}
	/**
	 * Get the level we are on.
	 */
	tpath = gtk_tree_model_get_path(model, &citer);
	depth = gtk_tree_path_get_depth(tpath);
	gtk_tree_path_free(tpath);
	/** Genre level */
	if(depth == 2)
	{
		GList *list, *node;
		gchar *genre;
		gtk_tree_model_get(model, &citer, PL3_CAT_INT_ID,&genre, -1);
		if(genre)
		{
			list = magnatune_db_get_url_list(genre, NULL ,NULL);
			for(node = list; node; node = node->next)
			{
				mpd_playlist_queue_add(connection, node->data);
			} 
			mpd_playlist_queue_commit(connection);
			g_list_foreach(list, (GFunc)g_free, NULL);		
			g_list_free(list);
			g_free(genre);
		}
	}	
	/** Artist */
	else if(depth == 3)
	{
		gchar *artist;
		gchar *genre;
		GtkTreeIter iter;
		if(gtk_tree_model_iter_parent(model, &iter, &citer))
		{
			GList *list, *node;
			gtk_tree_model_get(model, &iter, PL3_CAT_INT_ID,&genre, -1);
			gtk_tree_model_get(model, &citer, PL3_CAT_INT_ID,&artist, -1);
			list = magnatune_db_get_url_list(genre, artist,NULL);
			for(node = list; node; node = node->next)
			{
				mpd_playlist_queue_add(connection, node->data);
			} 
			mpd_playlist_queue_commit(connection);
			g_list_foreach(list, (GFunc)g_free, NULL);		
			g_list_free(list);
			g_free(genre);
			g_free(artist);
		}
	}




}

static void magnatune_cat_menu_replace()
{
	mpd_playlist_clear(connection);
	magnatune_cat_menu_add();
	mpd_player_play(connection);

}

static int magnatune_cat_menu_popup(GtkWidget *menu, int type, GtkWidget *tree, GdkEventButton *event)
{
	GtkWidget *item;
	if(type != plugin.id) return 0;
	else
    {
        /* add the clear widget */
        item = gtk_image_menu_item_new_from_stock(GTK_STOCK_REFRESH,NULL);
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
        g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(magnatune_redownload_reload_db), NULL);
        return 1;
    }
	return 0;
}

static void magnatune_add_selected(GtkWidget *button, GtkTreeView *tree)
{
	GtkTreeModel *model = GTK_TREE_MODEL(mt_store);
	GtkTreeSelection *sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
	GList *list = gtk_tree_selection_get_selected_rows(sel,&model);
	if(list)
	{
		GList *node ;
		for(node = list; node; node = g_list_next(node))
		{
			GtkTreeIter iter;
			if(gtk_tree_model_get_iter(model, &iter, node->data))
			{
                gchar *path;
				gtk_tree_model_get(model, &iter, MPDDATA_MODEL_COL_PATH, &path, -1);
                mpd_playlist_queue_add(connection, path);
                g_free(path);
            }
		}
		mpd_playlist_queue_commit(connection);
		g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
		g_list_free (list);
	}
}
static void magnatune_replace_selected(GtkWidget *button , GtkTreeView *tree)
{
	mpd_playlist_clear(connection);
	magnatune_add_selected(button, tree);
	mpd_player_play(connection);

}

static void magnatune_list_edit_columns(GtkMenuItem *item, GtkWidget *widget)
{
	gmpc_mpddata_treeview_edit_columns(GMPC_MPDDATA_TREEVIEW(widget));
}
static int magnatune_button_handle_release_event_tag_add(GtkWidget *button, gpointer userdata)
{
    GList *list;
    MpdData *data;
    int position = GPOINTER_TO_INT(userdata);
    gchar *genre=NULL, *artist=NULL, *album=NULL;
    GtkTreeSelection *select = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[0]));
    GtkTreeModel *model  = gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[0]));
    GtkTreeIter piter;
    if(gtk_tree_selection_get_selected(select, &model, &piter))
    {
        gtk_tree_model_get(model, &piter, MPDDATA_MODEL_COL_SONG_TITLE,&genre,-1);
    }
    if(position >= 1)
    {
        select = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[1]));
        model  = gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[1]));
        if(gtk_tree_selection_get_selected(select, &model, &piter))
        {
            gtk_tree_model_get(model, &piter, MPDDATA_MODEL_COL_SONG_TITLE,&artist,-1);
        }
    }
    if(position >= 2)
    {
        select = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeviews[2]));
        model  = gtk_tree_view_get_model(GTK_TREE_VIEW(treeviews[2]));
        if(gtk_tree_selection_get_selected(select, &model, &piter))
        {
            gtk_tree_model_get(model, &piter, MPDDATA_MODEL_COL_SONG_TITLE,&album,-1);
        }
    }
    list = magnatune_db_get_url_list(genre, artist, album);
    for(;list; list = list->next)
    {
        mpd_playlist_queue_add(connection, list->data);
    }
    mpd_playlist_queue_commit(connection);
    if(list)
    {
        g_list_foreach(list, g_free, NULL);
        g_list_free(list);
    }
    if(genre)
        g_free(genre);
    if(artist)
        g_free(artist);
    if(album)
        g_free(album);

}

static int magnatune_button_handle_release_event_tag_replace(GtkWidget *button, gpointer userdata)
{
    mpd_playlist_clear(connection);
    magnatune_button_handle_release_event_tag_add(button,userdata);
    mpd_player_play(connection);

}

static int magnatune_button_handle_release_event_tag(GtkWidget *tree, GdkEventButton *event,gpointer data)
{
    int i;
    int position = GPOINTER_TO_INT(data);
	if(event->button != 3) return FALSE;
	GtkWidget *item;

	GtkTreeSelection *sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
	if(gtk_tree_selection_count_selected_rows(sel)> 0)
	{
		GtkWidget *menu = gtk_menu_new();


        /* Add */
        item = gtk_image_menu_item_new_from_stock(GTK_STOCK_ADD,NULL);
		gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), item);
		g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK( magnatune_button_handle_release_event_tag_add), data);
		/* Replace */
		item = gtk_image_menu_item_new_with_label("Replace");
		gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item),
				gtk_image_new_from_stock(GTK_STOCK_REDO, GTK_ICON_SIZE_MENU));
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
        g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK( magnatune_button_handle_release_event_tag_replace), data);	
        gtk_widget_show_all(menu);
        gtk_menu_popup(GTK_MENU(menu), NULL, NULL,NULL, NULL, event->button, event->time);
        return TRUE;
    }
    return FALSE;
}

static int magnatune_button_release_event(GtkWidget *tree, GdkEventButton *event,gpointer data)
{
	if(event->button != 3) return FALSE;
	GtkWidget *item;

	GtkTreeSelection *sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));

	if(gtk_tree_selection_count_selected_rows(sel)> 0)
	{
		GtkWidget *menu = gtk_menu_new();
		GtkTreeModel *model = GTK_TREE_MODEL(mt_store);
		/* Add */
		item = gtk_image_menu_item_new_from_stock(GTK_STOCK_ADD,NULL);
		gtk_menu_shell_prepend(GTK_MENU_SHELL(menu), item);
		g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(magnatune_add_selected), tree);
		/* Replace */
		item = gtk_image_menu_item_new_with_label("Replace");
		gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item),
				gtk_image_new_from_stock(GTK_STOCK_REDO, GTK_ICON_SIZE_MENU));
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
		
		g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(magnatune_replace_selected), tree);
		
        /* Separator */	
        item = gtk_separator_menu_item_new();
        gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
        /* Edit columns */
        item = gtk_image_menu_item_new_with_label(("Edit Columns"));
		gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item),
				gtk_image_new_from_stock(GTK_STOCK_EDIT, GTK_ICON_SIZE_MENU));
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
		g_signal_connect(G_OBJECT(item), "activate",
				G_CALLBACK(magnatune_list_edit_columns), tree);

        gtk_widget_show_all(menu);
        
        gtk_menu_popup(GTK_MENU(menu), NULL, NULL,NULL, NULL, event->button, event->time);
        return TRUE;
	}
	return FALSE;
}

static void magnatune_cat_key_press(GtkWidget *tree, GdkEventKey *event, int selected_type)
{
	if(selected_type != plugin.id) return; 
	if(event->state&GDK_CONTROL_MASK && event->keyval == GDK_Insert)
	{
		magnatune_cat_menu_replace();
	}
	else if(event->keyval == GDK_Insert)
	{
		magnatune_cat_menu_add();	
	}

}

static int magnatune_key_press(GtkWidget *tree, GdkEventKey *event)
{
	if(event->state&GDK_CONTROL_MASK && event->keyval == GDK_Insert)
	{
		magnatune_replace_selected(NULL, GTK_TREE_VIEW(tree));
	}
	else if(event->keyval == GDK_Insert)
	{
		magnatune_add_selected(NULL, GTK_TREE_VIEW(tree));
	}
	else if (event->keyval == GDK_f && event->state&GDK_CONTROL_MASK)
	{
		treesearch_start(mt_tree_search);
		return TRUE;
	}
	else if((event->state&(GDK_CONTROL_MASK|GDK_MOD1_MASK)) == 0 && 
			((event->keyval >= GDK_space && event->keyval <= GDK_z)))
	{
		char data[2];
		data[0] = (char)gdk_keyval_to_unicode(event->keyval);
		data[1] = '\0';
		treesearch_start(TREESEARCH(mt_tree_search));
		gtk_entry_set_text(GTK_ENTRY(TREESEARCH(mt_tree_search)->entry),data);
		gtk_editable_set_position(GTK_EDITABLE(TREESEARCH(mt_tree_search)->entry),1);
		return TRUE;
	}
	return FALSE;
}
static void magnatune_save_myself(void)
{
    if(magnatune_vbox)
    {
        int position = gtk_paned_get_position(GTK_PANED(magnatune_vbox));
        cfg_set_single_value_as_int(config, "magnatune", "pane-pos", position);
    }
	if (magnatune_ref)
    {
        GtkTreePath *path = gtk_tree_row_reference_get_path(magnatune_ref);
        if(path)
        {
            gint *indices = gtk_tree_path_get_indices(path);
            debug_printf(DEBUG_INFO,"Saving myself to position: %i\n", indices[0]);
            cfg_set_single_value_as_int(config, "magnatune","position",indices[0]);
            gtk_tree_path_free(path);
        }
    }
}

static void magnatune_destroy(void)
{
    magnatune_db_destroy();
    if(magnatune_vbox)
        gtk_widget_destroy(magnatune_vbox);
}

gmpcMetaDataPlugin magnatune_cover = {
	magnatune_fetch_cover_priority,
	magnatune_fetch_get_image
};


/* Needed plugin_wp stuff */
gmpcPlBrowserPlugin magnatune_gbp = {
	.add                    = magnatune_add,
	.selected               = magnatune_selected,
	.unselected             = magnatune_unselected,
	.cat_right_mouse_menu   = magnatune_cat_menu_popup,
	.cat_key_press          = magnatune_cat_key_press
};

int plugin_api_version = PLUGIN_API_VERSION;

gmpcPlugin plugin = {
	.name                       = "Magnatune Stream Browser",
	.version                    = {PLUGIN_MAJOR_VERSION,PLUGIN_MINOR_VERSION,PLUGIN_MICRO_VERSION},
	.plugin_type                = GMPC_PLUGIN_PL_BROWSER|GMPC_PLUGIN_META_DATA,
    /* Creation and destruction */
	.init                       = magnatune_logo_init,
    .destroy                    = magnatune_destroy, 
    .save_yourself              = magnatune_save_myself,
    .mpd_status_changed         = magnatune_mpd_status_changed, 

    /* Browser extention */
	.browser                    = &magnatune_gbp,
	.metadata                   = &magnatune_cover,

    /* get/set enabled  */
	.get_enabled                = magnatune_get_enabled,
	.set_enabled                = magnatune_set_enabled
};


